package gregapi.item.multiitem.energy;

import static gregapi.data.CS.F;
import static gregapi.data.CS.NBT_ENERGY;
import static gregapi.data.CS.T;
import gregapi.code.HashSetNoNulls;
import gregapi.code.TagData;
import gregapi.item.IItemEnergy;
import gregapi.item.multiitem.MultiItem;
import gregapi.util.UT;

import java.util.Collection;

import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.inventory.IInventory;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.world.World;

public class EnergyStat implements IItemEnergy {
	public long mSize, mAmount, mCapacity;
	public TagData mType;
	public boolean mCanCharge, mCanDecharge;
	public ItemStack mEmptyItem, mHalfItem, mFullItem;
	
	public EnergyStat(TagData aEnergyType, long aCapacity, long aSize, long aAmount, boolean aCanCharge, boolean aCanDecharge, ItemStack aEmptyItem, ItemStack aHalfItem, ItemStack aFullItem) {
		mSize = aSize;
		mAmount = aAmount;
		mCapacity = aCapacity;
		mType = aEnergyType;
		mCanCharge = aCanCharge;
		mCanDecharge = aCanDecharge;
		mEmptyItem = aEmptyItem;
		mHalfItem = aHalfItem;
		mFullItem = aFullItem;
	}
	
	public static IItemEnergy makeREBattery(TagData aEnergyType, long aCapacity, long aSize, ItemStack aEmptyItem, ItemStack aHalfItem, ItemStack aFullItem) {
		return new EnergyStat(aEnergyType, aCapacity, aSize, 1, T, T, aEmptyItem, aHalfItem, aFullItem);
	}
	
	public static IItemEnergy makeSUBattery(TagData aEnergyType, long aCapacity, long aSize, ItemStack aEmptyItem, ItemStack aHalfItem, ItemStack aFullItem) {
		return new EnergyStat(aEnergyType, aCapacity, aSize, 1, F, T, aEmptyItem, aHalfItem, aFullItem);
	}
	
	public static IItemEnergy makeTool(TagData aEnergyType, long aCapacity, long aSize, ItemStack aEmptyItem, ItemStack aHalfItem, ItemStack aFullItem) {
		return new EnergyStat(aEnergyType, aCapacity, aSize, 1, T, F, aEmptyItem, aHalfItem, aFullItem);
	}
	
	public static IItemEnergy makeREBattery(TagData aEnergyType, long aCapacity, long aSize, ItemStack aItem) {
		return new EnergyStat(aEnergyType, aCapacity, aSize, 1, T, T, aItem, aItem, aItem);
	}
	
	public static IItemEnergy makeSUBattery(TagData aEnergyType, long aCapacity, long aSize, ItemStack aItem) {
		return new EnergyStat(aEnergyType, aCapacity, aSize, 1, F, T, aItem, aItem, aItem);
	}
	
	public static IItemEnergy makeTool(TagData aEnergyType, long aCapacity, long aSize, ItemStack aItem) {
		return new EnergyStat(aEnergyType, aCapacity, aSize, 1, T, F, aItem, aItem, aItem);
	}
	
	@Override
	public boolean isEnergyType(TagData aEnergyType, ItemStack aStack, boolean aEmitting) {
		return aEnergyType == mType;
	}
	
	@Override
	public long doEnergyInjection(TagData aEnergyType, ItemStack aStack, long aSize, long aAmount, IInventory aInventory, World aWorld, int aX, int aY, int aZ, boolean aDoInject) {
		aSize = Math.abs(aSize);
		if (!canEnergyInjection(aEnergyType, aStack, aSize)) return 0;
		long tStored = getEnergyStored(aEnergyType, aStack);
		if (tStored >= mCapacity) return 0;
		aAmount = Math.min(mAmount, aAmount);
		if (aAmount <= 0) return 0;
		long rAmount = 0;
		while (rAmount++<aAmount) if (tStored + rAmount * aSize >= mCapacity) break;
		if (aDoInject) setEnergyStored(aEnergyType, aStack, tStored + rAmount * aSize);
		return rAmount;
	}
	
	@Override
	public long doEnergyExtraction(TagData aEnergyType, ItemStack aStack, long aSize, long aAmount, IInventory aInventory, World aWorld, int aX, int aY, int aZ, boolean aDoExtract) {
		aSize = Math.abs(aSize);
		if (!canEnergyExtraction(aEnergyType, aStack, aSize)) return 0;
		long tStored = getEnergyStored(aEnergyType, aStack);
		if (tStored <= aSize) return 0;
		aAmount = Math.min(mAmount, aAmount);
		if (aAmount <= 0) return 0;
		long rAmount = 0;
		while (rAmount++<aAmount) if (tStored - rAmount * aSize < aSize) {rAmount--; break;}
		if (aDoExtract) setEnergyStored(aEnergyType, aStack, tStored - rAmount * aSize);
		return rAmount;
	}
	
	public ItemStack rechargeFromPlayer(TagData aEnergyType, ItemStack aStack, EntityLivingBase aPlayer, IInventory aInventory, World aWorld, int aX, int aY, int aZ) {
		if (aEnergyType != mType) return aStack;
		if (aStack.getItem() instanceof MultiItem) ((MultiItem)aStack.getItem()).chargeFromArmor(aStack, aPlayer);
		return aStack;
	}
	
	@Override
	public boolean useEnergy(TagData aEnergyType, ItemStack aStack, long aEnergyAmount, EntityLivingBase aPlayer, IInventory aInventory, World aWorld, int aX, int aY, int aZ, boolean aDoUse) {
		if (aPlayer instanceof EntityPlayer && ((EntityPlayer)aPlayer).capabilities.isCreativeMode) return T;
		if (aEnergyType != mType) return F;
		rechargeFromPlayer(aEnergyType, aStack, aPlayer, aInventory, aWorld, aX, aY, aZ);
		long tStored = getEnergyStored(aEnergyType, aStack);
		if (tStored >= aEnergyAmount) {
			if (aDoUse) {
				setEnergyStored(aEnergyType, aStack, tStored-aEnergyAmount);
				rechargeFromPlayer(aEnergyType, aStack, aPlayer, aInventory, aWorld, aX, aY, aZ);
			}
			return T;
		}
		if (aDoUse) {
			setEnergyStored(aEnergyType, aStack, 0);
			rechargeFromPlayer(aEnergyType, aStack, aPlayer, aInventory, aWorld, aX, aY, aZ);
		}
		return F;
	}
	
	@Override
	public ItemStack setEnergyStored(TagData aEnergyType, ItemStack aStack, long aAmount) {
		if (aEnergyType != mType) return aStack;
		
		NBTTagCompound tNBT = aStack.getTagCompound();
		if (tNBT == null) tNBT = new NBTTagCompound(); else tNBT.removeTag(NBT_ENERGY);
		
		aAmount = Math.min(aAmount, mCapacity);
		if (aAmount > 0) {
			if (aAmount >= mCapacity) {
				if (mFullItem != null) UT.Stacks.set(aStack, mFullItem, F, F);
			} else {
				if (mHalfItem != null) UT.Stacks.set(aStack, mHalfItem, F, F);
			}
			UT.NBT.setNumber(tNBT, NBT_ENERGY, aAmount);
		} else {
			if (mEmptyItem == null) {
				aStack.stackSize--;
			} else {
				UT.Stacks.set(aStack, mEmptyItem, F, F);
			}
		}
		UT.NBT.set(aStack, tNBT);
		return UT.Stacks.update_(aStack);
	}
	
	@Override
	public long getEnergyStored(TagData aEnergyType, ItemStack aStack) {
		if (aEnergyType != mType) return 0;
		NBTTagCompound tNBT = aStack.getTagCompound();
		return tNBT==null?0:tNBT.getLong(NBT_ENERGY);
	}
	
	@Override
	public long getEnergyCapacity(TagData aEnergyType, ItemStack aStack) {
		if (aEnergyType != mType) return 0;
		return mCapacity;
	}
	
	@Override
	public long getEnergySizeInputMin(TagData aEnergyType, ItemStack aStack) {
		if (aEnergyType != mType) return 0;
		return mSize <= 8 ? 1 : mSize / 2;
	}
	
	@Override
	public long getEnergySizeOutputMin(TagData aEnergyType, ItemStack aStack) {
		if (aEnergyType != mType) return 0;
		return mSize <= 8 ? 1 : mSize / 2;
	}
	
	@Override
	public long getEnergySizeInputRecommended(TagData aEnergyType, ItemStack aStack) {
		if (aEnergyType != mType) return 0;
		return mSize;
	}
	
	@Override
	public long getEnergySizeOutputRecommended(TagData aEnergyType, ItemStack aStack) {
		if (aEnergyType != mType) return 0;
		return mSize;
	}
	
	@Override
	public long getEnergySizeInputMax(TagData aEnergyType, ItemStack aStack) {
		if (aEnergyType != mType) return 0;
		return mSize * 2;
	}
	
	@Override
	public long getEnergySizeOutputMax(TagData aEnergyType, ItemStack aStack) {
		if (aEnergyType != mType) return 0;
		return mSize * 2;
	}
	
	@Override
	public Collection<TagData> getEnergyTypes(ItemStack aStack) {
		return new HashSetNoNulls<TagData>(F, mType);
	}
	
	@Override
	public boolean canEnergyInjection(TagData aEnergyType, ItemStack aStack, long aSize) {
		return aEnergyType == mType && aStack.stackSize == 1 && mCanCharge && aSize <= getEnergySizeInputMax(aEnergyType, aStack) && aSize >= getEnergySizeInputMin(aEnergyType, aStack);
	}
	
	@Override
	public boolean canEnergyExtraction(TagData aEnergyType, ItemStack aStack, long aSize) {
		return aEnergyType == mType && aStack.stackSize == 1 && mCanDecharge && aSize <= getEnergySizeOutputMax(aEnergyType, aStack) && aSize >= getEnergySizeOutputMin(aEnergyType, aStack);
	}
}