package gregapi.recipes.handlers;

import static gregapi.data.CS.*;
import gregapi.code.ArrayListNoNulls;
import gregapi.code.ICondition;
import gregapi.data.TD;
import gregapi.oredict.OreDictItemData;
import gregapi.oredict.OreDictMaterial;
import gregapi.oredict.OreDictPrefix;
import gregapi.recipes.IRecipeMapHandler;
import gregapi.recipes.Recipe.RecipeMap;
import gregapi.util.OM;
import gregapi.util.UT;
import net.minecraft.item.ItemStack;
import net.minecraftforge.fluids.Fluid;
import net.minecraftforge.fluids.FluidStack;

/**
 * @author Gregorius Techneticies
 */
public class RecipeMapHandlerPrefix implements IRecipeMapHandler {
	protected final ICondition mCondition;
	protected final OreDictPrefix[] mInputPrefixes, mOutputPrefixes;
	protected final byte[] mInputAmounts, mOutputAmounts;
	protected final ItemStack mAdditionalInput, mAdditionalOutput;
	protected final FluidStack mFluidInputPerUnit, mFluidOutputPerUnit;
	protected final long mUnitsInputted, mUnitsOutputted, mEUt, mDuration, mMultiplier;
	protected final boolean mAllowToGenerateAllRecipesAtOnce;
	
	protected boolean mAlreadyAddedAllRecipes = F, mOutputPulverizedRemains = F, mFlatFluidCosts = F;
	
	public RecipeMapHandlerPrefix(OreDictPrefix aInputPrefix, long aInputAmount, FluidStack aFluidInputPerUnit, long aEUt, long aDuration, long aMultiplier, FluidStack aFluidOutputPerUnit, OreDictPrefix aOutputPrefix, long aOutputAmount, ItemStack aAdditionalInput, ItemStack aAdditionalOutput, boolean aAllowToGenerateAllRecipesAtOnce, boolean aOutputPulverizedRemains, boolean aFlatFluidCosts, ICondition aCondition) {
		mAllowToGenerateAllRecipesAtOnce = aAllowToGenerateAllRecipesAtOnce;
		mOutputPulverizedRemains = aOutputPulverizedRemains;
		mFlatFluidCosts		= aFlatFluidCosts;
		mFluidInputPerUnit	= aFluidInputPerUnit;
		mFluidOutputPerUnit	= aFluidOutputPerUnit;
		mCondition			= aCondition;
		mInputPrefixes		= new OreDictPrefix[] {aInputPrefix};
		mInputAmounts		= new byte[] {UT.Code.bindStack(aInputAmount)};
		mOutputPrefixes		= aOutputPrefix == null ? ZL_OREDICTPREFIX : new OreDictPrefix[] {aOutputPrefix};
		mOutputAmounts		= new byte[] {UT.Code.bindStack(aOutputAmount)};
		mAdditionalOutput	= aAdditionalOutput;
		mAdditionalInput	= aAdditionalInput;
		mMultiplier			= aMultiplier;
		mDuration 			= aDuration;
		mEUt				= aEUt;
		long
		tUnitsProcessed		= 0;
		for (int i = 0; i < mInputPrefixes.length; i++) tUnitsProcessed += mInputPrefixes[i].mAmount * mInputAmounts[i];
		mUnitsInputted		= tUnitsProcessed;
		tUnitsProcessed		= 0;
		for (int i = 0; i < mOutputPrefixes.length; i++) tUnitsProcessed += mOutputPrefixes[i].mAmount * mOutputAmounts[i];
		mUnitsOutputted		= tUnitsProcessed;
	}
	
	public RecipeMapHandlerPrefix(OreDictPrefix aInputPrefix1, long aInputAmount1, OreDictPrefix aInputPrefix2, long aInputAmount2, FluidStack aFluidInputPerUnit, long aEUt, long aDuration, long aMultiplier, FluidStack aFluidOutputPerUnit, OreDictPrefix aOutputPrefix1, long aOutputAmount1, OreDictPrefix aOutputPrefix2, long aOutputAmount2, ItemStack aAdditionalInput, ItemStack aAdditionalOutput, boolean aAllowToGenerateAllRecipesAtOnce, boolean aOutputPulverizedRemains, boolean aFlatFluidCosts, ICondition aCondition) {
		mAllowToGenerateAllRecipesAtOnce = aAllowToGenerateAllRecipesAtOnce;
		mOutputPulverizedRemains = aOutputPulverizedRemains;
		mFlatFluidCosts		= aFlatFluidCosts;
		mFluidInputPerUnit	= aFluidInputPerUnit;
		mFluidOutputPerUnit	= aFluidOutputPerUnit;
		mCondition			= aCondition;
		mInputPrefixes		= aInputPrefix2 == null ? new OreDictPrefix[] {aInputPrefix1} : new OreDictPrefix[] {aInputPrefix1, aInputPrefix2};
		mInputAmounts		= aInputPrefix2 == null ? new byte[] {UT.Code.bindStack(aInputAmount1)} : new byte[] {UT.Code.bindStack(aInputAmount1), UT.Code.bindStack(aInputAmount2)};
		mOutputPrefixes		= aOutputPrefix2 == null ? aOutputPrefix1 == null ? ZL_OREDICTPREFIX : new OreDictPrefix[] {aOutputPrefix1} : new OreDictPrefix[] {aOutputPrefix1, aOutputPrefix2};
		mOutputAmounts		= aOutputPrefix2 == null ? new byte[] {UT.Code.bindStack(aOutputAmount1)} : new byte[] {UT.Code.bindStack(aOutputAmount1), UT.Code.bindStack(aOutputAmount2)};
		mAdditionalOutput	= aAdditionalOutput;
		mAdditionalInput	= aAdditionalInput;
		mMultiplier			= aMultiplier;
		mDuration 			= aDuration;
		mEUt				= aEUt;
		long
		tUnitsProcessed		= 0;
		for (int i = 0; i < mInputPrefixes.length; i++) tUnitsProcessed += mInputPrefixes[i].mAmount * mInputAmounts[i];
		mUnitsInputted		= tUnitsProcessed;
		tUnitsProcessed		= 0;
		for (int i = 0; i < mOutputPrefixes.length; i++) tUnitsProcessed += mOutputPrefixes[i].mAmount * mOutputAmounts[i];
		mUnitsOutputted		= tUnitsProcessed;
	}
	
	public RecipeMapHandlerPrefix(OreDictPrefix[] aInputPrefixes, long[] aInputAmount, FluidStack aFluidInputPerUnit, long aEUt, long aDuration, long aMultiplier, FluidStack aFluidOutputPerUnit, OreDictPrefix[] aOutputPrefixes, long[] aOutputAmount, ItemStack aAdditionalInput, ItemStack aAdditionalOutput, boolean aAllowToGenerateAllRecipesAtOnce, boolean aOutputPulverizedRemains, boolean aFlatFluidCosts, ICondition aCondition) {
		mAllowToGenerateAllRecipesAtOnce = aAllowToGenerateAllRecipesAtOnce;
		mOutputPulverizedRemains = aOutputPulverizedRemains;
		mFlatFluidCosts		= aFlatFluidCosts;
		mFluidInputPerUnit	= aFluidInputPerUnit;
		mFluidOutputPerUnit	= aFluidOutputPerUnit;
		mCondition			= aCondition;
		mInputPrefixes		= aInputPrefixes;
		mInputAmounts		= new byte[mInputPrefixes.length];
		for (int i = 0; i < mInputAmounts.length; i++) mInputAmounts[i] = UT.Code.bindStack(aInputAmount[i]);
		mOutputPrefixes		= aOutputPrefixes;
		mOutputAmounts		= new byte[mOutputPrefixes.length];
		for (int i = 0; i < mOutputAmounts.length; i++) mOutputAmounts[i] = UT.Code.bindStack(aOutputAmount[i]);
		mAdditionalOutput	= aAdditionalOutput;
		mAdditionalInput	= aAdditionalInput;
		mMultiplier			= aMultiplier;
		mDuration 			= aDuration;
		mEUt				= aEUt;
		long
		tUnitsProcessed		= 0;
		for (int i = 0; i < mInputPrefixes.length; i++) tUnitsProcessed += mInputPrefixes[i].mAmount * mInputAmounts[i];
		mUnitsInputted		= tUnitsProcessed;
		tUnitsProcessed		= 0;
		for (int i = 0; i < mOutputPrefixes.length; i++) tUnitsProcessed += mOutputPrefixes[i].mAmount * mOutputAmounts[i];
		mUnitsOutputted		= tUnitsProcessed;
	}
	
	@Override
	public boolean addRecipesUsing(RecipeMap aMap, ItemStack aStack, OreDictItemData aData) {
		if (isDone()) return F;
		if (UT.Stacks.equal(aStack, mAdditionalInput)) return mAllowToGenerateAllRecipesAtOnce && addAllRecipesInternal(aMap);
		return aData != null && aData.hasValidMaterialData() && UT.Code.contains(aData.mPrefix, mInputPrefixes) && addRecipeForMaterial(aMap, aData.mMaterial.mMaterial);
	}
	
	@Override
	public boolean addRecipesProducing(RecipeMap aMap, ItemStack aStack, OreDictItemData aData) {
		if (isDone()) return F;
		if (UT.Stacks.equal(aStack, mAdditionalOutput)) return mAllowToGenerateAllRecipesAtOnce && addAllRecipesInternal(aMap);
		return aData != null && aData.hasValidMaterialData() && UT.Code.contains(aData.mPrefix, mOutputPrefixes) && addRecipeForMaterial(aMap, aData.mMaterial.mMaterial);
	}
	
	@Override
	public boolean containsInput(RecipeMap aMap, ItemStack aStack, OreDictItemData aData) {
		if (isDone()) return F;
		return addRecipesUsing(aMap, aStack, aData);
	}
	
	@Override
	public boolean addRecipesUsing(RecipeMap aMap, Fluid aFluid) {
		if (isDone()) return F;
		return mFluidInputPerUnit != null && mFluidInputPerUnit.getFluid() == aFluid && mAllowToGenerateAllRecipesAtOnce && addAllRecipesInternal(aMap);
	}
	
	@Override
	public boolean addRecipesProducing(RecipeMap aMap, Fluid aFluid) {
		if (isDone()) return F;
		return mFluidOutputPerUnit != null && mFluidOutputPerUnit.getFluid() == aFluid && mAllowToGenerateAllRecipesAtOnce && addAllRecipesInternal(aMap);
	}
	
	@Override
	public boolean containsInput(RecipeMap aMap, Fluid aFluid) {
		return mFluidInputPerUnit != null && mFluidInputPerUnit.getFluid() == aFluid;
	}
	
	@Override
	public boolean addAllRecipes(RecipeMap aMap) {
		return mAllowToGenerateAllRecipesAtOnce && addAllRecipesInternal(aMap);
	}
	
	public boolean addAllRecipesInternal(RecipeMap aMap) {
		if (isDone()) return F;
		for (OreDictMaterial tMaterial : mInputPrefixes[0].mRegisteredMaterials) addRecipeForMaterial(aMap, tMaterial);
		mAlreadyAddedAllRecipes = T;
		return T;
	}
	
	@Override
	public boolean isDone() {
		return mAlreadyAddedAllRecipes;
	}
	
	public boolean addRecipeForMaterial(RecipeMap aMap, OreDictMaterial aMaterial) {
		if (!mCondition.isTrue(aMaterial) || aMaterial.contains(TD.Properties.INVALID_MATERIAL)) return F;
		
		ItemStack[] tInputs = new ItemStack[mAdditionalInput==null?mInputPrefixes.length:mInputPrefixes.length+1];
		if (mAdditionalInput != null) tInputs[tInputs.length-1] = mAdditionalInput;
		for (int i = 0; i < mInputPrefixes.length; i++) tInputs[i] = mInputPrefixes[i].mat(aMaterial, mInputAmounts[i]);
		for (ItemStack tInput : tInputs) if (UT.Stacks.invalid(tInput)) return F;
		
		ArrayListNoNulls<ItemStack> tOutputs = new ArrayListNoNulls<ItemStack>();
		OreDictMaterial tMaterial = getOutputMaterial(aMaterial);
		for (int i = 0; i < mOutputPrefixes.length; i++) tOutputs.add(mOutputPrefixes[i].mat(tMaterial, mOutputAmounts[i]));
		if (mOutputPulverizedRemains) tOutputs.add(OM.pulverize(aMaterial, mUnitsInputted-mUnitsOutputted));
		tOutputs.add(mAdditionalOutput);
		
		return !tOutputs.isEmpty() && aMap.addRecipeX(F,T,F,F,T, mEUt, mDuration<=0?getCosts(aMaterial):mDuration, tInputs, mFlatFluidCosts ? mFluidInputPerUnit : UT.Fluids.mul(mFluidInputPerUnit, aMaterial.mToolQuality+1), mFlatFluidCosts ? mFluidOutputPerUnit : UT.Fluids.mul(mFluidOutputPerUnit, aMaterial.mToolQuality+1), tOutputs.toArray(ZL_ITEMSTACK)) != null;
	}
	
	public OreDictMaterial getOutputMaterial(OreDictMaterial aMaterial) {
		return aMaterial;
	}
	
	public long getCosts(OreDictMaterial aMaterial) {
		return UT.Code.units(mUnitsInputted, U, mMultiplier+mMultiplier*aMaterial.mToolQuality, T);
	}
}